// Code generated by "stringer -type=EvalType"; DO NOT EDIT

package hil

import "fmt"

const (
	_EvalType_name_0 = "TypeInvalid"
	_EvalType_name_1 = "TypeString"
	_EvalType_name_2 = "TypeList"
	_EvalType_name_3 = "TypeMap"
)

var (
	_EvalType_index_0 = [...]uint8{0, 11}
	_EvalType_index_1 = [...]uint8{0, 10}
	_EvalType_index_2 = [...]uint8{0, 8}
	_EvalType_index_3 = [...]uint8{0, 7}
)

func (i EvalType) String() string {
	switch {
	case i == 0:
		return _EvalType_name_0
	case i == 2:
		return _EvalType_name_1
	case i == 4:
		return _EvalType_name_2
	case i == 8:
		return _EvalType_name_3
	default:
		return fmt.Sprintf("EvalType(%d)", i)
	}
}
